/*
	Generate a decent-looking square seal (kakuin) layout from
	a comma-separated string of kanji/kana, using whatever likely
	fonts are on installed on your system (worst-case, Meiryo and
	MS Mincho/Gothic).
*/

//TODO: add border radius (separate for inner and outer)
//TODO: add double border options
//TODO: add inner (column/row) borders

var opt = {
	font: "HakusyuTensyo",
	color: "red",
	text: "銀行印,落款印,之印",
	sizeh: 4, //inch
	sizew: 4, //inch
	border: 8, //pt
	margin: 4, //pt
	colspace: 6, //pt
	rowspace: 3, //pt
	roundOut: true,
	roundIn: true,
	equalWidth: true,
	negative: false,
	wariin: false,
	DEBUG: false,
};

//TODO nuke all pre-defined swatches and replace them with these
var colors = {
	white : [255,255,255],
    vermilion : [227,66,52],
    chineseRed : [170,56,30],
    black : [0,0,0]
};
for (color in colors) {
    var tmp = colors[color];
    colors[color]= new RGBColor();
    colors[color].red = tmp[0];
    colors[color].green = tmp[1];
    colors[color].blue = tmp[2];
}

var fonts = getfonts();

if (getoptions(opt)) {
	main();
}

function main() {
	var columns =[];
	var tmp = opt.text.split(/,/);
	for (i=0; i<tmp.length; i++) {
		columns.push({text: tmp[i]});
	}

	var preset = new DocumentPreset();
	if (File.fs == "Windows") {
		//why? because Windows CC turns on pixel alignment for a new RGB
		//document, and there's no way to override it in ExtendScript
		preset.colorMode = DocumentColorSpace.CMYK;
	}else{
		preset.colorMode = DocumentColorSpace.RGB;
	}
	preset.height = opt.sizeh;
	preset.width = opt.sizew;
	preset.rasterResolution = DocumentRasterResolution.HighResolution;
	preset.title = "hanko";
	preset.units =  RulerUnits.Points;

	var newdoc = app.documents.addDocument("hanko",preset);
	var layer = newdoc.layers[0];

	var centeredPara = newdoc.paragraphStyles.add("CenteredKanji");
	centeredPara.paragraphAttributes.justification = Justification.CENTER;

	makeborder(layer);

	if (opt.negative) {
		opt.color = colors.white;
	}

	var xLL = opt.border + opt.margin;
	var yLL = opt.border + opt.margin;
	var height = opt.sizeh - 2 * (opt.border + opt.margin);

	//create all text columns and calculate their sidebearings 
	//and dimensions at 100pt
	var maxLength = 0;
	for (i=0; i < columns.length; i++) {
		makecolumnstyle(newdoc,"Column " + (i+1),columns[i],opt.color);
		makecolumnframe(layer,centeredPara,columns[i]);
		realbounds(columns[i]);
		if (columns[i].text.length > maxLength) {
			maxLength = columns[i].text.length;
		}
	}

	//calculate column widths and resize each column to fit; y-axis scaling
	//is by font size, x-axis uses horizontalScale.
	var totalWidth = opt.sizew - 2 * (opt.border + opt.margin) 
			- (opt.colspace * (columns.length - 1));
	if (opt.equalWidth == true) {
		for (i=0; i < columns.length; i++) {
			columns[i].width = totalWidth / columns.length;
		}
	}else{
		//equal distortion
		var naturalWidth = 0;
		for (i=0; i < columns.length; i++) {
			columns[i].width =  totalWidth / columns[i].text.length;
			naturalWidth += columns[i].width;
		}
		for (i=0; i < columns.length; i++) {
			columns[i].width *= totalWidth / naturalWidth;
		}
	}

	debugbox(layer,xLL,yLL,opt.sizew - 2 * (opt.border + opt.margin),height);
	var x = xLL;
	for (j=0; j < columns.length; j++) {
		var i = columns.length - j - 1;
		var y = yLL;
		var xscale = columns[i].width / columns[i].bounds.width;
		var yscale = height / columns[i].bounds.height;
		columns[i].style.size *= yscale;
		columns[i].style.characterAttributes.horizontalScale = 
			100 * xscale / yscale;
	//	debugbox(layer,x,y,columns[i].bounds.width * xscale,height);
		columns[i].frame.position = [x - columns[i].bounds.dx * xscale,
			height + y - columns[i].bounds.dy * yscale];
		splitcolumn(columns[i],yscale);

		x += columns[i].width + opt.colspace;
		if (i > 0) {
			debugbox(layer,x-opt.colspace,y,opt.colspace,height);
		}
	}
}

function debugbox(layer,left,top,width,height) {
	if (opt.DEBUG) {
		var tmp = layer.pathItems.rectangle(top,left,width,-1 * height);
		tmp.filled = false;
		tmp.strokeColor = colors.black;
		tmp.strokeWidth = 0.1;
		tmp.strokeDashes = [3,3];
	}
}

//calculates real per-character bounding box with opt.rowspace
//instead of normal vertical spacing (to pack Reisho better).
//Spacing is based on a sample character (曜/ボ) that is pretty
//square, to keep everything the same size; this does force you
//to do manual resizing in some fonts.
function realbounds(column) {
	var dupe = column.frame.duplicate();
	var tmplength = dupe.contents.length;
	if (iskana(dupe.contents)) {
		dupe.contents = "ボ";
	}else{
		dupe.contents = "曜";
	}
	var origBounds = dupe.geometricBounds;
	var outline = dupe.createOutline();
	var realBounds = outline.geometricBounds;
	column.bounds = {
		dx: realBounds[0] - origBounds[0],
		dy: realBounds[1] - origBounds[1],
		width: outline.width,
		height: outline.height * tmplength + (tmplength - 1) * opt.rowspace,
		charHeight: outline.height
	};
	outline.remove();
}

//take a properly sized and positioned column of text, and break
//it up into one row per character, so it can be positioned without
//excess vertical space (e.g. Reisho)
function splitcolumn(column,yscale) {
	var text = column.frame.contents.split("");
	var y = column.frame.top;
	for (var i=0; i< text.length; i++) {
		var dupe = column.frame.duplicate();
		dupe.contents = text[i];
		dupe.top = y;
		y -= (column.bounds.charHeight + opt.rowspace) * yscale;
	}
	column.frame.remove();
}

function makecolumnstyle(doc,name,column,color) {
	column.style = doc.characterStyles.add(name);
	var tmp = column.style.characterAttributes;
	tmp.textFont = textFonts.getByName(opt.font);
	tmp.size = 100;
	tmp.fillColor = color;
}

function makecolumnframe(layer,para,column) {
	column.frame = layer.textFrames.add();
	column.frame.pixelAligned = false;
	column.frame.orientation = TextOrientation.VERTICAL;
	column.frame.opticalAlignment = true;
	column.frame.contents = column.text;
	para.applyTo(column.frame.paragraphs[0]);
	column.style.applyTo(column.frame.textRange);
	//for some reason, the font from the character style
	//is not applied, so we do it manually
	column.frame.textRange.characterAttributes.textFont = 
		textFonts.getByName(opt.font);
}

function makeborder(layer) {
	var path = layer.compoundPathItems.add();
	var item;
	if (opt.wariin == true) {
		item = makeSquashedRectangle(path,0);
	}else if (opt.roundOut == true) {
		item = path.pathItems.roundedRectangle(0,0,opt.sizew,-1 * opt.sizeh,
			5,5,true);
	}else{
		item = path.pathItems.rectangle(0,0,opt.sizew,-1 * opt.sizeh);
	}
	item.pixelAligned = false;
	item.filled = true;
	item.stroked = false;
	item.fillColor = opt.color;
	item.evenodd = true;
	if (opt.negative) {
		return(1);
	}
	if (opt.wariin == true) {
		item = makeSquashedRectangle(path,opt.border);
	} else if (opt.roundIn == true) {
		item = path.pathItems.roundedRectangle(opt.border,opt.border,
			opt.sizew-opt.border*2,-1 * (opt.sizeh - opt.border*2),5,5,true);
		item.evenodd = true;
	}else{
		item = path.pathItems.rectangle(opt.border,opt.border,
			opt.sizew-opt.border*2,-1 * (opt.sizeh - opt.border*2),true);
		item.evenodd = true;
	}
	item.pixelAligned = false;
}

function makeSquashedRectangle(path,offset) {
	var item;
	var xdelta = opt.sizew/2;
	var ydelta = opt.sizew*1/3;
	var xdelta2 = xdelta/2 * (1 + offset/ydelta);
	var ydelta2 = ydelta/2 * (1 + offset/xdelta);
	item = path.pathItems.add();
	var pt1 = item.pathPoints.add();
	pt1.anchor = [offset,opt.sizeh-ydelta];
	pt1.leftDirection = pt1.anchor;
	pt1.rightDirection = [offset,opt.sizeh-ydelta2];
	pt1.pointType = PointType.SMOOTH;
	var pt2 = item.pathPoints.add();
	pt2.anchor = [xdelta,opt.sizeh-offset];
	pt2.leftDirection = [xdelta2,opt.sizeh-offset];
	pt2.rightDirection = [opt.sizew-xdelta2,opt.sizeh-offset];
	pt2.pointType = PointType.SMOOTH;
	var pt3 = item.pathPoints.add();
	pt3.anchor = [opt.sizew-offset,opt.sizeh-ydelta];
	pt3.leftDirection = [opt.sizew-offset,opt.sizeh-ydelta2];
	pt3.rightDirection = pt3.anchor;
	pt3.pointType = PointType.SMOOTH;
	var pt4 = item.pathPoints.add();
	pt4.anchor = [opt.sizew-offset,ydelta];
	pt4.leftDirection = pt4.anchor;
	pt4.rightDirection = [opt.sizew-offset,ydelta2];
	pt4.pointType = PointType.SMOOTH;
	var pt5 = item.pathPoints.add();
	pt5.anchor = [xdelta,offset];
	pt5.leftDirection = [opt.sizew-xdelta2,offset];
	pt5.rightDirection = [xdelta2,offset];
	pt5.pointType = PointType.SMOOTH;
	var pt6 = item.pathPoints.add();
	pt6.anchor = [offset,ydelta];
	pt6.leftDirection = [offset,ydelta2];
	pt6.rightDirection = pt6.anchor;
	pt6.pointType = PointType.SMOOTH;
	item.closed = true;
	return item;
}

function iskana(s) {
	for (var i=0; i<s.length; i++) {
		var c = s.charCodeAt(i);
		if (c < 0x3040 || c > 0x30FF) {
			return(false);
		}
	}
	return(true);
}

//TODO: pull the font regexp up to the top to make it easier to change
function getfonts() {
	var nfonts = textFonts.length;
	var fonts = [];
	for(var i=0; i<nfonts; i++) {
 		if (textFonts[i].name.match(/^(D[CF]|HG|Hakusyu|Hkusyu|Meiryo|MS-(Mi|Go))/)) {
			fonts.push(textFonts[i].name);
		}else if (textFonts[i].name.match(/-W\d+$/)) {
			fonts.push(textFonts[i].name);
		}
	}
	fonts.sort();
	return fonts;
}

function getoptions(opt) {
	var optWindow = new Window("dialog","Seal Options");
	optWindow.alignChildren = "left";

	var g1 = optWindow.add("group");
	var color = g1.add("dropdownlist",undefined,["red","black"]);
	color.selection = 0;

	var font = g1.add("dropdownlist",undefined,fonts);
	font.maximumSize.height = optWindow.maximumSize.height / 2;
	font.selection = 0;
	for (var i=0; i<fonts.length; i++) {
		if (fonts[i] == opt.font) {
			font.selection = i;
		}
	}

	var manualfont = g1.add("edittext",undefined,"(other; use PostScript name)");
	manualfont.preferredSize.width = 16*12;

	var g2 = optWindow.add("group");
	g2.alignChildren = ["left","top"];
	var text = g2.add("edittext",undefined,opt.text);
	text.graphics.font = "dialog:24";
	text.preferredSize.width = 24 * 20;
	text.active = true;

	var gc = optWindow.add("group");
	var g3 = gc.add("panel");
	g3.orientation = "column";
	g3.alignChildren= ["right","top"];

	var g3a = g3.add("group");
	g3a.add("statictext",undefined,"Seal height:");
	var sizeh = g3a.add("edittext",undefined,opt.sizeh);
	sizeh.characters = 4;
	var unith = g3a.add("dropdownlist",undefined,["in","mm","cm","pt"]);
	unith.selection = 0;

	var g3b = g3.add("group");
	g3b.add("statictext",undefined,"Seal width:");
	var sizew = g3b.add("edittext",undefined,opt.sizew);
	sizew.characters = 4;
	var unitw = g3b.add("dropdownlist",undefined,["in","mm","cm","pt"]);
	unitw.selection = 0;

	var g3c = g3.add("group");
	g3c.add("statictext",undefined,"Border:");
	var borderWidth = g3c.add("edittext",undefined,opt.border);
	borderWidth.characters = 4;
	var unitb = g3c.add("dropdownlist",undefined,["in","mm","cm","pt"]);
	unitb.selection = 3;

	var g3d = g3.add("group");
	g3d.add("statictext",undefined,"Margin:");
	var margin = g3d.add("edittext",undefined,opt.margin);
	margin.characters = 4;
	var unitm = g3d.add("dropdownlist",undefined,["in","mm","cm","pt"]);
	unitm.selection = 3;

	var g3e = g3.add("group");
	g3e.add("statictext",undefined,"Col Spacing:");
	var colspace = g3e.add("edittext",undefined,opt.colspace);
	colspace.characters = 4;
	var unitc = g3e.add("dropdownlist",undefined,["in","mm","cm","pt"]);
	unitc.selection = 3;

	var g3f = g3.add("group");
	g3f.add("statictext",undefined,"Row Spacing:");
	var rowspace = g3f.add("edittext",undefined,opt.rowspace);
	rowspace.characters = 4;
	var unitr = g3f.add("dropdownlist",undefined,["in","mm","cm","pt"]);
	unitr.selection = 3;

	var g4 = gc.add("group");
	g4.alignment = "top";
	g4.alignChildren = ["left","top"];
	g4.orientation = "column";
	var roundOut = g4.add("checkbox",undefined,"Round outer corners?");
	roundOut.value = opt.roundOut;
	var roundIn = g4.add("checkbox",undefined,"Round inner corners?");
	roundIn.value = opt.roundIn;
	var equalWidth = g4.add("checkbox",undefined,"Equal-width columns?");
	equalWidth.value = opt.equalWidth;
	var negative = g4.add("checkbox",undefined,"Invert colors?");
	negative.value = opt.negative;
	var wariin = g4.add("checkbox",undefined,"Squashed rectangle?");
	wariin.value = opt.wariin;
	var DEBUG = g4.add("checkbox",undefined,"DEBUG");
	DEBUG.value = opt.DEBUG;

	var g4a = g4.add("group");
	g4a.alignment = "right";
	g4a.add("button",undefined,"Cancel");
	g4a.add("button",undefined,"OK");

	if (File.fs == "Windows") {
		//why? because Windows CC sizes windows wrong on HiDPI displays
		g4a.alignment = "left";
		optWindow.onShow = function () {
			optWindow.size[0]*= 1.5;
			optWindow.size[1] *= 1.5;
			optWindow.location[0] /= 2;
			optWindow.location[1] /= 2;
		};
	}
	if (optWindow.show() == 1) {
		opt.DEBUG = DEBUG.value;
		opt.sizeh = UnitValue(sizeh.text,unith.selection.text).as("px");
		opt.sizew = UnitValue(sizew.text,unitw.selection.text).as("px");
		opt.font = font.selection.text;
		if (manualfont.text != "(other; use PostScript name)") {
			opt.font = manualfont.text;
		}
		//TODO: fix color selection
		if (color.selection.text == "red") {
			opt.color = colors.chineseRed;
		}else{
			opt.color = colors.black;
		}
		opt.text = text.text;
		opt.border = UnitValue(borderWidth.text,unitb.selection.text).as("px");
		opt.margin = UnitValue(margin.text,unitm.selection.text).as("px");
		opt.colspace = UnitValue(colspace.text,unitc.selection.text).as("px");
		opt.rowspace = UnitValue(rowspace.text,unitr.selection.text).as("px");
		opt.roundOut = roundOut.value;
		opt.roundIn = roundIn.value;
		opt.equalWidth = equalWidth.value;
		opt.negative = negative.value;
		opt.wariin = wariin.value;
		return true;
	}else{
		return false;
	}
}
